/*
 * wine browser linker
 *
 * Launches the required unix browser
 *
 * Copyright (c) TransGaming Technologies 2003
 *
 * David Hammerton <david@transgaming.com>
 *
 * parts take from sims installer and sims browser proxy.
*/

#define WINE_NOWINSOCK
#include <windows.h>
#undef WINE_NOWINSOCK

/* unix includes (i hope) */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/wait.h>



typedef struct
{
    char *name;
    char *filename;
    char *exec;
} browser_data;

/* list shamelessly (mostly) stolen from sims installer */
static const browser_data init_browser_data[] =
{
      { "Mozilla (/usr/bin)", "/usr/bin/mozilla", "/usr/bin/mozilla %s" },
      { "Mozilla (/usr/local/bin)", "/usr/local/bin/mozilla", "/usr/local/bin/mozilla %s" },
      { "Konqueror (/usr/bin)", "/usr/bin/konqueror", "/usr/bin/konqueror %s" },
      { "Konqueror (/usr/local/bin)", "/usr/local/bin/konqueror", "/usr/local/bin/konqueror %s" },
      { "Galeon (/usr/bin)", "/usr/bin/galeon", "/usr/bin/galeon %s" },
      { "Galeon (/usr/local/bin)", "/usr/local/bin/galeon", "/usr/local/bin/galeon %s" },
      { "Mozilla-Firebird (/usr/bin)", "/usr/bin/mozilla-firebird", "/usr/bin/mozilla-firebird %s" },
      { "Mozilla-Firebird (/usr/local/bin)", "/usr/local/bin/mozilla-firebird", "/usr/local/bin/mozilla-firebird %s" },
      { "Netscape (/usr/bin)", "/usr/bin/netscape", "/usr/bin/netscape %s" },
      { "Netscape (/usr/X11R6/bin)", "/usr/X11R6/bin/netscape", "/usr/X11R6/bin/netscape %s" },
      { "Netscape (/usr/local/bin)", "/usr/local/bin/netscape", "/usr/local/bin/netscape %s" },
      { NULL, NULL }
};

int get_working_builtin_browser()
{
    int i;
    struct stat fs;
    for (i = 0; init_browser_data[i].name != NULL; i++)
    {
        if (stat(init_browser_data[i].filename, &fs) == 0) /* 0 === success */
            return i;
    }
    return -1;
}

BOOL get_browser_line(char *buffer, int buffer_len)
{
    HKEY hkey;
    buffer[0] = 0;
    char *tmp;
    int formatcount;

    /* this stuff kinda taken from shellink.c */
    /* need to get the winebrowser (eg, web browser used by wine */
    if (!GetEnvironmentVariableA("WINEBROWSER", buffer, buffer_len))
    {
        /* otherwise get from registry. the config file puts its entry
         * into the registry */
        if (!RegOpenKeyExA(HKEY_LOCAL_MACHINE, "Software\\Wine\\Wine\\Config\\Wine",
                           0, KEY_ALL_ACCESS, &hkey))
        {
            DWORD type, count = buffer_len;
            if (RegQueryValueExA(hkey, "Browser", 0, &type, buffer, &count)) buffer[0] = 0;
            RegCloseKey(hkey);
        }
    }
    if (!*buffer)
    {
        int i;
        fprintf(stderr, "WARNING: no browser entry found in config, will try to find the best match\n");
        if ((i = get_working_builtin_browser()) == -1)
            return FALSE;
        if (strlen(init_browser_data[i].exec) > buffer_len) return FALSE;
        strncpy(buffer, init_browser_data[i].exec, buffer_len);
    }

    tmp = buffer;
    formatcount = 0;
    while ((tmp = strchr(tmp, '%')))
    {
        /* remove all format chars apart from the first %s */
        if (formatcount || tmp[1] != 's')
        {
            tmp[0] = '\\';
        }
        else formatcount++;
        tmp++;
    }
    if (!formatcount)
    {
        /* append a " %s" onto the end of the browser line */
        const char *append = " %s";
        if (buffer_len < (strlen(buffer) + strlen(append) + 1))
            return FALSE;
        strcat(buffer, append);
    }

    return TRUE;
}

/* child death will be handled in wait*. */
void setup_sigchld()
{
    /* The default action is to ignore SIGCHLD. */
    struct sigaction sa;

    sa.sa_handler = SIG_DFL;
    sigemptyset(&sa.sa_mask);
    sigaddset(&sa.sa_mask, SIGSTOP);
    sigaddset(&sa.sa_mask, SIGKILL);
    sa.sa_flags = SA_NOCLDSTOP | SA_RESTART;

    sigaction(SIGCHLD, &sa, NULL);
}


BOOL fork_exec_and_wait(char *browser_fmt, char *url)
{
    pid_t child;
    char buffer[MAX_PATH];
    snprintf(buffer, sizeof(buffer), browser_fmt, url);
    int status;

    setup_sigchld();
    child = fork();
    if (child == -1)
    {
        return FALSE;
    }
    if (child == 0)
    { /* child */
        fprintf(stderr, "executing browser line: '%s'\n", buffer);
        unsetenv("TEMP"); /* mozilla doesn't like wines TEMP (or TMP) env variable */
        unsetenv("TMP");
        system(buffer); /* easier than exec* */
        _exit(1);
    }
    /* parent */

    waitpid(child, &status, 0);

    return TRUE;
}

int CALLBACK WinMain (HINSTANCE inst, HINSTANCE prev, LPSTR cmdline, int show)
{
    char buffer[MAX_PATH];
    BOOL ret;

    ret = get_browser_line(buffer, sizeof(buffer));
    if (!ret)
    {
        fprintf(stderr, "ERROR: failed to find a browser. couldn't launch url '%s'\n",
                        cmdline);
        return 0;
    }

    ret = fork_exec_and_wait(buffer, cmdline);
    if (!ret)
    {
        fprintf(stderr, "ERROR: failed execute your browser (%s). couldn't launch url '%s'\n",
                        buffer, cmdline);
        return 0;
    }

    return 1;
}

